#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#include "control.h"
#include "graphics.h"
#include "callback.h"
#include "fitsigma.h"
#include "simu.h"
#include "calc.h"
#include "gvar.h"

extern Container c[];

/*------------------------------------------------------------------*/
/*------------------------------------------------------------------*/

/*
 * Calculate atom initial coordinates according to system parameters.
 */
void CalcAtoms()
{
  int i;
    
  if (r != NULL)  free(r);
  if (atype != NULL) free(atype);
  if (borderatom != NULL) free(borderatom);
  if (fixvel != NULL) free(fixvel);

  if (/*boundaryFlag*/1) 
    r = FormBoundary(size.x,size.y,Pi*seam.rotL/180.0,Pi*seam.rotR/180.0,
		     lParam,lParam2,seam.disp.x,seam.disp.y,&nAtom);
  else {
    r = CalcRect(size.x, size.y-seam.downcuta, Pi * seam.rotL / 180.0, lParam, &nAtom);
    if ((atype = (int*)malloc((nAtom) * sizeof(int))) == NULL) {
      printf("Error allocating memory for atype!\n");
      bexit(-1);
    }
    for(i=0;i<nAtom;i++) atype[i]=1;
  }
  
  if (crackFlag) {
    CreateInitialCrack(r,&nAtom,crack.center.x,crack.center.y,crack.size.x,
		       crack.size.y,crack.angle);
    if ((r = realloc(r, (nAtom+ADDITION) * sizeof(VecD))) == NULL) {
      printf("Error allocating memory\n");
      bexit(-1);
    }
  }
  if (verticalFlag) {
    /* fix the vertical edges */ 
    for (i = 0; i < nAtom; i++) 
      if (fabs(r[i].x) > size.x / 2.0 - borderW) 
	atype[i]=-(atype[i]);
    /* fix the bottom */
  } else 
    for (i = 0; i < nAtom; i++) 
      if (r[i].y < -(size.y / 2.0 - borderW)) 
	atype[i]=-(atype[i]);

  /* check for border atoms */

  if ((borderatom = (int*)malloc((nAtom) * sizeof(int))) == NULL) {
    printf("Error allocating memory for borderatom!\n");
    bexit(-1);
  }

  for (i = 0; i < nAtom; i++) {
    borderatom[i]=0;
    fixvel[i]=0;
    cforce[i].x=0.0;
    cforce[i].y=0.0;
    if (r[i].x < -size.x/2.0+borderW) borderatom[i] |= BORDER_LEFT;
    if (r[i].x >  size.x/2.0-borderW) borderatom[i] |= BORDER_RIGHT;
    if (r[i].y < -size.y/2.0+borderW) borderatom[i] |= BORDER_BOTTOM;
    if (r[i].y >  size.y/2.0-borderW) borderatom[i] |= BORDER_TOP;
  }
}


/*-----------------------------------------------------------*/
/*-----------------------------------------------------------*/ 


/*
 * calculate atom coordinates for a rectangular (1 1 1) plane of size lx x ly
 */
VecD *CalcRect(double lx, double ly, double angle, double a, int *atoms)
{
  int i, j, n, iMin, iMax, jMax, nMax;
  double rd, x, y, x1, x2, y2;
  double tmpmin,tmpmax;
  VecD *r;

  /* radius of a circular disk drawn around the rectangular area */

  rd = sqrt(SQ(lx) + SQ(ly));
  x1 = a / sqrt(2.0); /* distance between two atoms in x-directional row */
  x2 = cos(Pi / 3.0) * x1; /* x-distance between atoms in subsequent 
			      rows = 1/2 */
  y2 = sin(Pi / 3.0) * x1; /* y-distance between atoms in subsequent 
			      rows = sqrt(3)/2*/

  /* allocate memory for atom coordinates */

  nMax = (int)(ceil(lx / x1) * ceil(ly / y2));
  tmpmin=1.0e30;
  tmpmax=-1.0e30;

  if ((r = (VecD*)malloc((nMax+ADDITION) * sizeof(VecD))) == NULL) {
    fprintf(stderr,"Error allocating memory\n");
    bexit(-1);
  }
  if ((fixvel = (int*)malloc((nMax+ADDITION) * sizeof(int))) == NULL) {
    fprintf(stderr,"Error allocating memory\n");
    bexit(-1);
  }
  n = 0; /* number of atoms */
  jMax = (int)(rd / y2);
  for (j = -jMax; j <= jMax; j++) { /* vertical loop */
    iMin = (int)(-sqrt(SQ(rd) - SQ((double)j * y2)) / x1 -(double)(j/2));
    iMax = (int)(sqrt(SQ(rd) - SQ((double)j * y2)) / x1 -(double)(j/2));
    for (i = iMin; i <= iMax; i++) { /* horizontal loop */
      if (n >= nMax)
	break;
      x = (double)i * x1 + (double)j * x2 +0.5; /* calculate coordinates */
      y = (double)j * y2;
      r[n].x = cos(angle) * x - sin(angle) * y; /* do rotation */
      r[n].y = cos(angle) * y + sin(angle) * x;
    
      /* reject atoms that are outside the rectangular area */
      
      if (r[n].x>-lx/2.0 && r[n].x<=lx/2.0 && r[n].y>-ly/2.0 
	  && r[n].y<=ly/2.0) {
	if (r[n].x<tmpmin) tmpmin=r[n].x;
	if (r[n].x>tmpmax) tmpmax=r[n].x;
	++n;
      }
    }
  }
  
  /* free unused memory */
  
  if ((r = realloc(r, (n+ADDITION) * sizeof(VecD))) == NULL) {
    printf("Error allocating memory\n");
    bexit(-1);
  }
  *atoms = n;
  return r;
}



/*------------------------------------------------------------------*/
/*------------------------------------------------------------------*/


/*
 * calculate atom coordinates for two (1 1 1) planes separated by dx, dy
 */

VecD *FormBoundary(double lx, double ly, double angle1, double angle2,
		   double a1,double a2, double dx, double dy, int *atoms)
{
  
  double boudist,boudistfix,Anear,Bnear;/* These parameters are used when 
					   calculating the minimum distance
					   between atoms in the horizontal 
					   boundary */
  double laycount;
  double rel;                 /*lattice parameter for atomtype 2*/
  int i, l;
  int n,u,d;
  
  VecD *r, *rtemp;
/*
  rtemp = CalcRect(lx, ly+ ceil(*//*2.0 *//* fabs(dy))+seam.Amove, 
		   angle1, a1, &n);
  */
  if (verticalFlag)
    rtemp = CalcRect(lx-seam.downcuta, ly + ceil(/*2.0 */ fabs(dy)),
		     angle1, a1, &n);
  else
    rtemp = CalcRect(lx, ly-seam.downcuta+ ceil(/*2.0 */ fabs(dy))+seam.Amove, 
		     angle1, a1, &n);
  if ((r = (VecD*)malloc((n + ADDITION) * sizeof(VecD))) == NULL) {
    printf("Error allocating memory\n");
    bexit(-1);
  }
  if ((cforce = (VecD*)malloc((n + ADDITION) * sizeof(VecD))) == NULL) {
    printf("Error allocating memory\n");
    bexit(-1);
  }
  if ((fixvel = (int*)malloc((n + ADDITION) * sizeof(int))) == NULL) {
    printf("Error allocating memory\n");
    bexit(-1);
  }
  if ((atype = (int*)malloc((n + ADDITION) * sizeof(int))) == NULL) {
    printf("Error allocating memory\n");
    bexit(-1);
  }
  
  Anear=-ly/2;
  Bnear=ly/2;
  laycount=-ly/2;
  u=0;
  d=0;
 /* 
  for (i = 0; i < n+ADDITION; i++) {
    fixvel[i]=0;
    cforce.x=0.0;
    cforce.y=0.0;
  }*/
    
  /* Vertical Boundary */
  if (verticalFlag) {
    for (i = 0, l = 0; i < n; i++) {
      if (rtemp[i].x > -((lx - dx) / 2.0-seam.downcuta) 
	  && rtemp[i].x <(seam.boundpos+seam.upcuta)) {
	r[l].x = rtemp[i].x - dx / 2.0+seam.Amove;
	r[l].y = rtemp[i].y;
	atype[l]=1;
	/*fixvel[l]=0;*/
	l++;
      }
      
    }
    free(rtemp);
    /*right side */
    if(a2<a1){
      rel=a1-((a1-a2)/100)*relaxation;
    }
    else
      rel =a1+((a2-a1)/100)*relaxation;
    rtemp = CalcRect(lx+seam.upcut + ceil(2.0 * fabs(dx)), ly + ceil(fabs(dy))
		     -seam.dot,angle2, rel, &n);
    for (i = 0; i < n; i++) {
      if (l > n + ADDITION)
	break;
      r[l].y = rtemp[i].y + dy; 
      if (rtemp[i].x < ((lx - dx) / 2.0+seam.upcut) && 
	  rtemp[i].x >=(seam.boundpos+seam.downcut) && fabs(r[l].y) < ly / 2.0)
	{
	  r[l].x = rtemp[i].x + dx / 2.0+seam.Bmove;
	  
	  if (boundaryFlag)
	    atype[l]=2;
	  else
	    atype[l]=1;
	  l++;}
    }
  }
  
  /* Horisontal Boundary */
  /* bottom side */ 
 
  else {
    for (i = 0, l = 0; i < n; i++) {
      if (rtemp[i].y > (-(ly - dy) / 2.0+seam.downcuta) && 
	  rtemp[i].y <(seam.boundpos+seam.upcuta)){
	r[l].y = rtemp[i].y - dy / 2.0+seam.Amove;
	r[l].x = rtemp[i].x;
	/*fixvel[l]=0;*/
	if(r[l].y>Anear){
	  d++;
	  Anear=r[l].y;
	}
	atype[l]=1;
	l++;
      }
    }
    free(rtemp);
    /*Upper side*/
    
    if(a2<a1){
      rel=a1-((a1-a2)/100)*relaxation;
    }
    else
      rel =a1+((a2-a1)/100)*relaxation;
    rtemp = CalcRect(lx+ ceil(2.0*fabs(dx)), ly + ceil( fabs(dy))+seam.upcut,
		     angle2, rel, &n);
    for (i = 0; i < n; i++) {
      if (l > n + ADDITION)
	break;
      r[l].x = rtemp[i].x + dx;

      /*check if the atom is in the lattice area */ 
      if (rtemp[i].y < ((ly - dy) / 2.0+seam.upcut)  && 
	  rtemp[i].y >=(seam.boundpos+seam.downcut) &&  
	  fabs(r[l].x) < lx / 2.0) {
	int take = 0;
	/*if an island of atoms b on top of atoms a is wanted then*/ 
	if(seam.dot !=0){
	  /*check if atom in islands area*/
	  if(r[l].x > -(lx/2.0-dx-seam.dot) &&  r[l].x < 
	     (lx/2.0+dx-seam.dot)){
	    r[l].y = rtemp[i].y + dy / 2.0+seam.Bmove;
	    take = 1;
	  }
	}
	/*both layers are as wide*/
	else{
	  r[l].y = rtemp[i].y + dy / 2.0+seam.Bmove;
	  take = 1;
	}
	
	if(r[l].y<Bnear){
	  Bnear=r[l].y;
	}
	if(r[l].y>laycount){
	  laycount=r[l].y;
	  u++;
	}
	if(take){
	  if (boundaryFlag)
	    atype[l]=2;
	  else
	    atype[l]=1;
	l++;	
	}
      }
    }
    
    boudist=Bnear-Anear;
    boudistfix=sin(Pi/3.0)*a1/sqrt(2.0)-boudist;
    if (!runningFlag && suggestFlag && !verticalFlag)
      printf("Fixing suggestion: y=%f\n",boudistfix);
  }
  free(rtemp);
  /* free unused memory */
  if ((r = realloc(r, (l+ADDITION) * sizeof(VecD))) == NULL) {
    printf("Error allocating memory\n");
    bexit(-1);
  }
  if ((cforce = realloc(cforce, (l+ADDITION) * sizeof(VecD))) == NULL) {
    printf("Error allocating memory\n");
    bexit(-1);
  }
  if ((atype = realloc(atype, (l+ADDITION) * sizeof(int))) == NULL) {
    printf("Error allocating memory\n");
    bexit(-1);
  }
   if ((fixvel = realloc(fixvel, (l+ADDITION) * sizeof(int))) == NULL) {
    printf("Error allocating memory\n");
    bexit(-1);
  }
  *atoms = l;
  return r;
}



/*------------------------------------------------------------------*/
/*------------------------------------------------------------------*/


/* Create a crack */
void CreateInitialCrack(VecD *r, int *nAtom, double x, double y,
			double sizex, double sizey, double angle)
{
  int i;
  double dx,dy;
  
  angle *= Pi / 180.0;

  for (i = 0; i < *nAtom; ++i) {
    dx=(r[i].x-x)*cos(angle)+(r[i].y-y)*sin(angle);
    dy=(r[i].y-y)*cos(angle)-(r[i].x-x)*sin(angle);
    if (fabs(dx) < 0.5*sizex && fabs(dy) < 0.5*sizey) {
      (*nAtom)--;
      r[i].x = r[*nAtom].x;
      r[i].y = r[*nAtom].y;
      atype[i]=atype[*nAtom];
      fixvel[i]=fixvel[*nAtom];
      --i;
    }
  }
  /*
  if ((r = realloc(r, (*nAtom+ADDITION) * sizeof(VecD))) == NULL) {
    printf("Error allocating memory\n");
    bexit(-1);
  }*/
}


/*------------------------------------------------------------------*/
/*------------------------------------------------------------------*/




/* 
   This is ugly but it works in most cases.
   Adjust simulation area in x direction
   to be an integer number of atomic spacings
   if we have periodic boundary conditions 
   This does not work if the system is rotated!!
*/

void adjustsimarea(int pbc,double *sarea, double latticea) 
{
  double tmp1,tmp2,ax;
  if (pbc) {
    ax=latticea/sqrt(2.0);
    tmp1 = ax * ( floor((*sarea)/ax) );
    tmp2 = ax * ( floor((*sarea)/ax) + 1 );
    if (*sarea-tmp1<tmp2-*sarea) *sarea=tmp1;
    else *sarea=tmp2;
  }
}

/*---------------------------------------------------------*/
/*---------------------------------------------------------*/

